/* converted from h8dinfo.c  Aug 18 2016 HRJ */
/*  H8dinfo.c Written & Compiled with Software Toolworks C/80 V. 3.0 under HDOS by
** Glenn Roberts 10 November 2013
** from SEBHC discussion group download 11/14/13
*/

#include <stdio.h>
#include <stdlib.h> /* for malloc, etc. HRJ*/
#include <conio.h> /* for cgets */

#define	sint		unsigned short
#define	TRUE		1
#define	FALSE		0

#define	NOTHING	0	/* choices for getfile() */
#define	DUMP		1
#define	TXT			2
#define	EXTRT		3
FILE *chan; 			/* write file I/O channel */

#define	SECSIZE	256
#define MAXSECT 399		/* 10 X 40 sectors */
#define	MAXD		64	/* number of dir entries */
/* two groups of 4 physical sectors, each with 8 entries */

#define	ESIZE		32	/* CP/M directory item size 	*/

#define EMPTY 0xE5		/* CP/M fill for unwritten or formatted sector */

#pragma pack(push, 1)			//HRJ forces byte alignment on structures

/* struct dkinfo {} dkbuff; doesn't exist in CP/M

/* CP/M file directory entry data structure */
struct cpminfo {
	char erase;		/* E5h? if erased */
	char name[8];	/* file name		*/
	char ext[3];	/* file extension	*/
	char count;		/* count of entries for filename	*/
	char ver;		/*  (not used)	*/
	char stt;		/*  (not used)	*/
	char clu;		/* # of 128 byte clusters */
	char gr[16];		/* 16 group values 02-FF hex or 00 */
	// end of CPM directory entry, rest computed in program */
	sint size;	/* file size (sectors)*/
	sint group[16];	/* program converts groups to sints */
	char dname[13];	/* name in string form w/null */
} cpmentry;

#pragma pack(pop)	//HRJ restore previous alignment see previous pragma

/* array of pointers to directory entries */
struct cpminfo *direntries[MAXD];

unsigned char rwbuff[512];	/* disk I/O buffer	*/
char grtbuff[SECSIZE];		/* GRT buffer		*/
#define INBUF 32			/* number of input chars */
char strbuf[INBUF+5];		/* Console I/O buffer	*/
FILE *channel; 			/* file I/O channel	*/
int nentries;			/* number of directory entries found */
int map;			/*true or false to shop group mapping */

/* logphy - convert logical sector to physical sector
** physical	0 1 2 3 4 5 6 7 8 9
** logical	0 5 3 8 1 6 4 9 2 7
*/
int logphy[] = {0, 4, 8, 2, 6, 1, 5, 9, 3, 7};


/* readsectors - read one or more sectors (n) on the disk
** image starting at the ith sector.  The sectors are
** numbered 0, 1, ... 399.
** Uses global file handle "channel".  The calling program
** must ensure that buff will hold the result.
*/
void readsectors(char *buff, int i, int n)
{
	int err;

	if (n > MAXSECT) {
		printf("attempt to read sector %i beyond 100K disk image\n",n);
		exit(2); /* bad sector */
	}
	/* seek() from beginning of file by sector-size */
	err = fseek(channel, (long int)i*SECSIZE, SEEK_SET);
	if (err != 0) {
		printf("error reading disk image sector %i/n",i);
		exit(1); /* seek error*/
	}

	/* in position, now read the sectors... */
	//read(channel, buff, n*SECSIZE);
	fread(buff,1,n*SECSIZE,channel);
}

/* cpminfo - see if file a CP/M H8D file. Returns
** 1 if error encountered or 0 if no error
*/
int cpminfo(void)
{
	int rc = FALSE;

	/* Test if H8D format by examining the first few
	** bytes of sector 0 for a proper signature (several varieties).
	*/
	readsectors(rwbuff,0,1);

	if ( ( rwbuff[0]==0xC3 && rwbuff[1]==0xA0 && rwbuff[2]==0x22 && rwbuff[3]==0x20 ) ||
	     ( rwbuff[0]==0xAF && rwbuff[1]==0xD3 && rwbuff[2]==0x7D && rwbuff[3]==0xCD ) ) {
		/* have H8D file */
		printf ("Has HDOS signature on sector 0\n");
		rc = FALSE;
	}

	else {
		readsectors(rwbuff,3 * 10,1); /* track 3 sector 0 is directory */
		if ( isprnt(rwbuff[1]) && isprnt(rwbuff[2]) && isprnt(rwbuff[3]) ) {
		/* one printable directory filename entry */
		printf ("Input file has printable CPM directory filename\n");
		rc = TRUE;
		}
	}

	return rc;
}

/* builddir - read CP/M system directory and populate
** directory array.  we need to start with first dir sector
** and read all the entries available
** through eight directory sectors max.
*/

void builddir(void)
{
	int i, j, cc, done, lsect, psect, nxtsector;
	unsigned clu;
	struct cpminfo *entry;
	unsigned char *src, *dst, *sdir, *ddir;
	int *inxt;
	sint *sz;

	nxtsector = 10 * 3; /*track 3 sector 0*/
	lsect = 0; psect = logphy[lsect];
	done = FALSE;
	nentries = 0; /* number of directory entries */
	printf("Building directory...");
	do {
		/* read a sector of the directory track */
		readsectors(rwbuff, nxtsector + psect,1);

		/* each of these phy sectors holds 8 directory
		** entries. loop over them and store them
		** in the array (allocating space as we go)
		*/
		src = rwbuff;
		for (i=0; ((i<8) && (!done)); i++) {
			if ((*src) != (char)EMPTY ){
				/* attempt to allocate an entry */
				if (nentries == MAXD) {
					printf("Error - directory full!\n");
					done = TRUE;
				}
				// else if((entry = sbrk(sizeof(cpmentry))) == -1)
				else if((entry = malloc(sizeof(cpmentry))) == NULL) {
					printf("Error allocating directory entry!\n");
					done = TRUE;
				}
				else {
					/* copy ESIZE bytes from the buffer into the entry */
					dst = (char *) entry;
					for (j=0; j<ESIZE; j++)
						*dst++ = *src++;

					/* store link to it in the directory array */
					direntries[nentries] = entry;

					/* fill out full file name in direntries */
					sdir = direntries[nentries]->name;
					ddir = direntries[nentries]->dname;
					for (j=0; ((j<8) && (*sdir!=0) && (*sdir!=0x20)); j++)
						*ddir++ = *sdir++;

					sdir = direntries[nentries]->ext;
					if (*sdir != 0) {
						*ddir++ = '.';
						for (j=0; ((j<3) && (*sdir!=0) && (*sdir!=0x20)); j++) {
							*sdir = (*sdir) & 0x7f; /* clears MSB in extent name*/
							*ddir++ = *sdir++;
						}
					}
					*ddir = 0;

					/* read cluster size and convert to size */
					direntries[nentries]->size = (unsigned int) direntries[nentries]->clu;
					if (direntries[nentries]->size > 128) direntries[nentries]->size = 128;

					/* read groups and convert to sint's */
					for (j=0; (j<16); j++)
						direntries[nentries]->group[j] = (unsigned int) direntries[nentries]->gr[j];

					nentries++;
				}
			} else {
				/* blank entry, skip over */
				src += ESIZE;
			}
		}
		/* now get the next sector of directory */
		if (!done) {
			lsect++; if (lsect >= 8) done = TRUE; /* only 8 directory sectors */
			psect = logphy[lsect];
		}
	} while (!done);
	printf("Done! %d directory entries found\n", nentries);
}

/* listdir - print directory listing from
** stored array.
*/
void listdir(void)
{
	int i, j, entries, siztot;
	char *c;
	int first, next, last;

	entries = 0;
	siztot = 0;
	/* list directory */
	printf("filename.ext        size  extent   groups \n");
	for (i=0; i<nentries; i++) {
		c = direntries[i]->name;
		if (isprnt(*c)) {
			++entries;
			/* print name and extension */
			for (j=0; j<8; j++, c++)
				putchar((*c==0) ? ' ' : *c);
			putchar('.');
			c = direntries[i]->ext;
			for (j=0; j<3; j++, c++)
				putchar((*c==0) ? ' ' : *c);

			/* print flags */
			if (direntries[i]->erase != (char) 0) printf(" erase ");
			else printf ("       ");

			/* print file size in blocks, accumulate total */
			siztot += direntries[i]->size;
			printf(" %4d  ", direntries[i]->size);

			/* print number of directory entry */
			printf(" %2d ", direntries[i]->count);

			/* print first and last groups */
			first = last = direntries[i]->group[0];
			for (j=0; j<16; j++) {
				next = direntries[i]->group[j];
				if (next > last) last= next;
			}
			printf(" %4d to %4d ",first, last);

			printf("\n");
		}
	}
	printf("Files %d, Total %d\n", entries, siztot);
}


static const char h[10][80]= {
"H8D CPM disk 10 sect per track. CPM on trk 0,1,2, dir on trk 3\n\0",
"A CPM directory entry is an extent, up to 16 groups per filename.\n\0",
"Each group is 8 CPM blocks of 128 bytes,\n\0",
"Contained in 4 physical sectors of 256 bytes so there may be excess.\n\0",
"An extent of 16 groups is 16 X 8 = 128 blocks = 0x80.\n\0",
"Larger files use multiple extent directory entries by filename.\n\0",
"The 10 physical sectors are mapped into 10 logical sectors by array logphy[].\n\0",
"So a track holds 10 phy sectors or 2.5 groups, in LOGICAL sector order.\n\0",
"Most CPM text files end with 0x1A but binary files may or may not.\n\0",
"Expect garbage at the last block of CPM files.\n\0"
};

/* printinfo - print information on CPM file */
void printinfo(void) {

	int i;
	for (i=0; i<10; i++) printf("%s",h[i]);

}


/* match - match file name to directory entry and count.
** Returns the index of the file name in the
** directory array, or -1 if not found
** fname format "filename.c\0" no spaces
*/
int match(char *fname, int cnt)
{
	int i, rc, match;

	rc = -1;
	// printf("Looking for match to: %s entry %i\n", fname, cnt);
	/* loop over directory entries looking for a match */
	for (i=0, match=FALSE; (i<nentries) && (!match); i++)
		if (strcmp(fname, direntries[i]->dname) == 0) {
			if (direntries[i]->count == cnt){
				match = TRUE;
				rc = i;
			}
		}

	return rc;
}

/* findeof - find 0x1a end of file mark in buffer
*/
int findeof(char *buff) {
	int i;

	for (i=0; i<SECSIZE; i++) {
		if (buff[i] == (char) 0x1a) return (i);
	}
	return (SECSIZE+1);

}



/* getfile - dump the contents of a file by directory extent, choose dump in hex, text or write.
*/
void getfile(char *fname, int ext, int choice)
{
	int ifile, i, j, sc, done, abssec;
	int startsec, trk, logsec, physec, blocks, count, wrsize;
	unsigned grp;

	if ((ifile=match(fname,ext)) == -1)
		printf("File %s extent %i not found!\n", fname, ext);
	else {
		printf("Getting file %s extent %i\n", direntries[ifile]->dname, ext);
		blocks = direntries[ifile]->size;
		printf("File extent %i size in blocks %i\n",ext, blocks); /* #of 128 byte blocks */

		/* read thru group list and dump the file
		** for all sectors in each group
		*/
		done = FALSE; wrsize = SECSIZE;
		for (i=0; (i<16 && !done); i++) {
			grp = direntries[ifile]->group[i];
			if (grp <= 0) done = TRUE;
			else if (grp > 128) {
				done = TRUE;
				printf ("bad group number found\n");
			}
			else { /* good group number */
				startsec = (grp*4) + 30; 	/* group 0 is track 3 sec 0*/
				trk = startsec/10;		/* integer division */
				logsec = startsec - trk*10;	/* remainder */
				for (j=0;(j<=3 && !done);j++) { /* up to 4 sectors in group */
					physec = logphy[logsec];
					if (map) printf("\nmap: g=%i, t=%i, l=%i, p=%i, bks=%i",grp,trk,logsec,physec,blocks);
					readsectors(rwbuff, (trk*10 + physec),1); /* read sector */

					/* if dump sector in hex */

					if (choice == DUMP) {
						// printf("dump\n");
						hexdump(rwbuff,SECSIZE);
					}

					blocks-=2; /* whole sector is two CPM blocks in file */
					/* if blocks is 0 or -1 we are done, if lower a problem */
					if (blocks == -1) { /* only first half of sector in file */
						printf("\ntrack %i sector %i, first 128 bytes ends file, EOF at %3XH\n",trk,physec, findeof(rwbuff));
						wrsize = 128; /* could set to eof byte but let's be prudent */
						done = TRUE;
					}
					else if (blocks == 0) { /* sector completes file*/
						printf("\ntrack %i sector %i, all 256 bytes ends file, EOF at %3XH\n",trk,physec, findeof(rwbuff));
						wrsize = SECSIZE;
						done = TRUE;
					}
					else if (blocks < -1) { /* negative block value, error */
						printf("\nbad block count =%i\n",blocks);
						done = TRUE;
					}
					/* next sector, to next track if necessary */
					logsec++; if (logsec>9) {
						trk++; logsec=0;
					}

					/* write sector to file if extracting */
					if (choice == EXTRT)  {
						count = fwrite(rwbuff, 1, wrsize, chan);
						if (count != wrsize)
							printf("Write error in file %s\n", fname);
						/* show progress... */
						putchar('.');
					}

				}

			}
		}
	}
}

/* dumpfile - dump the contents of a file across multiple directory extents */
void dumpfile(char *fname)
{
	int ifile, ext, i, done;

	ext = 0;
	if ((ifile=match(fname,ext)) == -1) { /* does filename appear once? */
		printf("File %s not found!\n", fname);
		return;
	}
	/* at least one entry, find more */
	done = FALSE;
	for (i=1; i<MAXD && !done; i++) if ((ifile=match(fname,i)) > -1) ext++; else done = TRUE;

	/* go thru directory extents for filename */
	for (i=0; i<=ext; i++) {
		// printf("obtaining sectors for directory extent %i\n",i);
		getfile(fname, i, DUMP); /* will hex dump file */
	}

}

void extract(char *fname) {
	int ifile, ext, i, done, choice;

	ext = 0;
	if ((ifile=match(fname,ext)) == -1) { /* does filename appear once? */
		printf("File %s not found!\n", fname);
		return;
	}
	/* at least one entry, find more */
	done = FALSE;
	for (i=1; i<MAXD && !done; i++) if ((ifile=match(fname,i)) > -1) ext++; else done = TRUE;

	/* open output file for binary write */
	if ((chan = fopen(fname, "wb")) == 0) {
		printf("Error opening destination file %s\n", fname);
		return;
	}
	else printf("Extracting file %s\n",fname);

	/* go thru directory extents for filename */
	for (i=0; i<=ext; i++) {
		// printf("obtaining sectors for directory extent %i\n",i);
		getfile(fname, i, EXTRT); /* will write to file */
	}
	/* close output file */
	fclose(chan); printf ("\n");

}



/* hexline - print n bytes in hexadecimal */
void hexline(b,n)
char *b;
int n;
{
  int i;

  for (i=0; i<n; i++)
    printf(" %02x",b[i]&0xFF);
}

/* asciiline - print n characters in ascii */
/* (substitute '.' for non printable characters) */
void asciiline(b,n)
char *b;
int n;
{
  int i;

  for (i=0; i<n; i++) {
    if (isprnt(b[i]))
      putchar(b[i]);
    else
      putchar('.');
  }
}



/* hexdump - dump a buffer to stdout in hexadecimal format */
void hexdump(b,n)
char *b;
int n;
{
  int i;

  /* print header */
  printf("\n\t0  1  2  3  4  5  6  7  8  9  A  B  C  D  E  F");
  printf("  0123456789ABCDEF\n");
  printf("\t\b-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- --");
  printf("  ----------------\n");
  for (i=0; i<n; i+=16) {
    printf("%04x",i);
    printf(" :");
    hexline(b+i,16);
    printf("  ");
    asciiline(b+i,16);
    printf("\n");
  }
}

/* HRJ was isprint but this is not C standard isprint */
int isprnt(c)
char c;
{
  return ((c>0x1F) && (c<0x7F));
}

void strupper(s)
char *s;
{
	char *c;

	for (c=s; *c; c++)
		*c = toupper(*c);
}


int main(argc, argv)
int argc;
char *argv[];
{
	int done, secno;


	printf("H8D CPM Sector Dump and Analysis Utility\n\n");

	if (argc < 2)
		printf("h8dcpm usage: %s <file>\n", argv[0]);
	else if((channel = fopen(argv[1], "rb")) == 0)
		printf("Unable to open source file %s\n", argv[1]);
	/* attempt to read disk for HDOS or CP/M signatures */
	else if(!cpminfo())
		printf("File %s not a valid CPM file!\n", argv[1]);
	else {
		/* valid file - build and populate directory array */
		builddir();

		done = FALSE; map = TRUE;
		while (!done) {
			printf("Command: I)nfo  H)elp  F)ile  D)ir  S)ector  M)ap eX)tract  Q)uit\n");
			strbuf[0]=INBUF; cgets(strbuf);
			switch(toupper(strbuf[2])) {
			case 'I':
				printinfo();
			break;
			case 'H':
				printf("Commands:\n\n");
				printf("\tInfo:\tDisplay key info on CPM file\n");
				printf("\tHelp:\tDisplay help\n");
				printf("\tFile:\tDisplay hex/text dump of a file\n");
				printf("\tDir:\tList the CPM disk file directory\n");
				printf("\tSector:\tdump a specified Sector\n");
				printf("\teXtract:\tExtract a file to this system\n");
				printf("\tMap: \ttoggle Map display of groups and logical, physical sectors\n");
				printf("\tQuit: \tExit\n\n");
			break;
			case 'D':
				listdir();
			break;
			case 'F':
				printf("File to display? : ");
				strbuf[0]=INBUF; cgets(strbuf);
				strupper(strbuf+2);
				dumpfile(strbuf+2);
			break;
			case 'X':
				printf("File to extract? : ");
				strbuf[0]=INBUF; cgets(strbuf);
				strupper(strbuf+2);
				extract(strbuf+2);
			break;
			case 'S':
				printf("Which Sector ?   : ");
				strbuf[0]=INBUF; cgets(strbuf);
				secno = atoi(strbuf+2);
				if (secno <= MAXSECT) {
					readsectors(rwbuff,secno,1);
					hexdump(rwbuff,SECSIZE);
				}
				else
					printf("Illegal sector\n");
			break;
			case 'Q':
				printf("Exiting...\n");
				done = TRUE;
			break;
			case 'M':
				printf("Map display of groups and logical, physical sectors");
				if (map == TRUE) map = FALSE; else map = TRUE;
				if (map) printf(" is ON\n"); else printf (" is OFF\n");
			break;
			}
		}
		fclose(channel);
	}
}

